/*

   makesram V1.0 for UNIX by madman
   This is a UNIX version of LaC's makesram util for uploading
   SRAM files to a Cart SRAM.  The .v64 file is his work, not mine,
   this is merely a UNIX version of his code, which was quite easy
   to figure out.

   ROM Checksum Routine Courtesy of:
   chksum64 V1.2, a program to calculate the ROM checksum of Nintendo64 ROMs.
   Copyright (C) 1997  Andreas Sterbenz (stan@sbox.tu-graz.ac.at)
*/

#include <stdio.h>

//#define max2(a, b) ( (a)>(b) ? (a) : (b) )
//#define min2(a, b) ( (a)<(b) ? (a) : (b) )

#define BUFSIZE 32768

#define CHECKSUM_START 0x1000
#define CHECKSUM_LENGTH 0x100000L
#define CHECKSUM_HEADERPOS 0x10
#define CHECKSUM_END (CHECKSUM_START + CHECKSUM_LENGTH)

#define CHECKSUM_STARTVALUE 0xf8ca4ddc

#define ROL(i, b) (((i)<<(b)) | ((i)>>(32-(b))))
/*
#define BYTES2LONG(b, s) ( (((b)[0^(s)] & 0xffL) << 24) | \
                           (((b)[1^(s)] & 0xffL) << 16) | \
                           (((b)[2^(s)] & 0xffL) <<  8) | \
                           (((b)[3^(s)] & 0xffL)) )

#define LONG2BYTES(l, b, s)  (b)[0^(s)] = ((l)>>24)&0xff; \
                             (b)[1^(s)] = ((l)>>16)&0xff; \
                             (b)[2^(s)] = ((l)>> 8)&0xff; \
                             (b)[3^(s)] = ((l)    )&0xff;
*/
#define SRAM_BEGIN 0x0286C0

void makesram(void)
{
  FILE *file_ptr, *file_ptr2;
  unsigned char buffer1[BUFSIZE];
  long fsize;
  unsigned long sum1, sum2;
//  int swapped = -1;
//  int readonly = 0;
  char sram[511], tempchar;
  int i = 0;


  if ((file_ptr = fopen (cmd1, "r+")) == NULL)
    {
      printf ("Error: %s not found.\n", cmd1);
      exit (1);
    }
  if ((file_ptr2 = fopen (cmd2, "r")) == NULL)
    {
      printf ("Error: %s not found.\n", cmd2);
      exit (1);
    }

  fseek (file_ptr2, 0, SEEK_END);
  fsize = ftell (file_ptr2);
  if (fsize != 512)
    {
      printf ("Error: SRAM File is not 512 bytes!\n");
      exit (1);
    }

  fseek (file_ptr, 0, SEEK_END);
  fsize = ftell (file_ptr);
  if (fsize != 1310720)
    {
      printf ("Error: ROM File is not proper size! (Check filename)\n");
      exit (1);
    }

  fseek (file_ptr2, 0, SEEK_SET);
  fread (sram, sizeof (sram), 1, file_ptr2);

  while (i < 511)
    {
      tempchar = sram[i];
      sram[i] = sram[i + 1];
      sram[i + 1] = tempchar;
      i += 2;
    }
  fseek (file_ptr, SRAM_BEGIN, SEEK_SET);
  fwrite (sram, sizeof (sram), 1, file_ptr);
  printf ("SRAM Written...\n");
  fseek (file_ptr, CHECKSUM_START, SEEK_SET);

  {
    unsigned long i;
    unsigned long c1, k1, k2;
    unsigned long t1, t2, t3, t4;
    unsigned long t5, t6;
    unsigned int n;
    long clen = CHECKSUM_LENGTH;
    long rlen = fsize - CHECKSUM_START;

/*     Below is the actual checksum calculation algorithm, which was
       reverse engineered out of Nagra's program.

       As you can see, the algorithm is total crap. Obviously it was
       designed to be difficult to guess and reverse engineer, and not
       to give a good checksum. A simple XOR and ROL 13 would give a
       just as good checksum. The ifs and the data dependent ROL are really
       extreme nonsense.
*/
    t1 = CHECKSUM_STARTVALUE;
    t2 = CHECKSUM_STARTVALUE;
    t3 = CHECKSUM_STARTVALUE;
    t4 = CHECKSUM_STARTVALUE;
    t5 = CHECKSUM_STARTVALUE;
    t6 = CHECKSUM_STARTVALUE;

    for (;;)
      {
	if (rlen > 0)
	  {
	    n = fread (buffer1, 1, min2 (BUFSIZE, clen), file_ptr);
	    if ((n & 0x03) != 0)
	      {
		n += fread (buffer1 + n, 1, 4 - (n & 3), file_ptr);
	      }
	  }
	else
	  {
	    n = min2 (BUFSIZE, clen);
	  }
	if ((n == 0) || ((n & 3) != 0))
	  break;
	for (i = 0; i < n; i += 4)
	  {
	    c1 = BYTES2LONG (&buffer1[i], 1);
	    k1 = t6 + c1;
	    if (k1 < t6)
	      t4++;
	    t6 = k1;
	    t3 ^= c1;
	    k2 = c1 & 0x1f;
	    k1 = ROL (c1, k2);
	    t5 += k1;
	    if (c1 < t2)
	      {
		t2 ^= k1;
	      }
	    else
	      {
		t2 ^= t6 ^ c1;
	      }
	    t1 += c1 ^ t5;
	  }
	if (rlen > 0)
	  {
	    rlen -= n;
	    if (rlen <= 0)
	      memset (buffer1, 0, BUFSIZE);
	  }
	clen -= n;
      }
    sum1 = t6 ^ t4 ^ t3;
    sum2 = t5 ^ t2 ^ t1;
  }
  LONG2BYTES (sum1, &buffer1[0], 0);
  LONG2BYTES (sum2, &buffer1[4], 0);
  fseek (file_ptr, CHECKSUM_HEADERPOS, SEEK_SET);
  fread (buffer1 + 8, 1, 8, file_ptr);
  fseek (file_ptr, CHECKSUM_HEADERPOS, SEEK_SET);
  LONG2BYTES (sum1, &buffer1[16], 1);
  LONG2BYTES (sum2, &buffer1[20], 1);
  if (fwrite (buffer1 + 16, 1, 8, file_ptr) != 8)
    {
      printf ("Error: Could not write new checksum to ROM!\n");
      exit (1);
    }

  printf ("Checksum Written...\nFile now updated with new SRAM.\n");
  fclose (file_ptr);
  fclose (file_ptr2);

}
