#include "global.h"
#include "interact.h"
#include "v64proto.h"
#include "parsys.h"     /* definition of 'system-dependent' functions */

  char buffer1[BUFSIZE];
  int biosVersion;
	time_t start;
	unsigned long flen=0;

int mode;

#ifdef LINUX
void spperror(void)
{
	printf("Try switching the port back to SPP (Standart, Normal) mode in your bios.\n");
}
#endif

void status(long bs,long fl)
{
long cps=0;
time_t curr,left;
char buf[127],buf2[20];

if(!flen)
{
	flen=fl;
}
curr=time(0)-start;

if(curr!=0)
{
	cps=bs/curr;
	left=(flen-bs)/cps;


	strftime(buf2,11,"%M:%S",gmtime(&left));
	strcpy(buf,"\
===================");
buf[20*bs/flen]='\0';
strcat(buf,"\
-------------------");
buf[20]='\0';



	printf("\r%8ld bytes sent. [%s], CPS=%lu, ETA=%s.   ", bs, buf,cps,buf2);
}
}

int guessByteOrder(unsigned char rombytes[12])
{
  if( BYTES2LONG(rombytes, 0) == HEADER_MAGIC ) {
    return 0;
  } else if( BYTES2LONG(rombytes, 1) == HEADER_MAGIC ) {
    return 1;
  }
  if( (rombytes[8] == 0x80) && (rombytes[9] != 0x80) ) {
    return 2;
  } else {
    if( rombytes[9] == 0x80 ) {
      return 3;
    }
  }
  return -1;
}

int sendBytesFromFile(int baseport, FILE *file1, long flen1, int v64format)
{
  long bytessent=0;
  int sendmem=0;

  printf("%8ld bytes (%ld.%02ld Mbit) will be transmitted.\n", flen1, flen1>>17, ((flen1*100)>>17)%100);

  fflush(stdout);
  for( ;; ) {
    unsigned int n;

    if( sendmem == 0 ) {
      n = fread(buffer1, 1, min2(flen1, BUFSIZE), file1);
    }
    if( (sendmem != 0) || (n == 0) ) {
      if( sendmem == 0 ) {
        sendmem = 1;
        memset(buffer1, 0, BUFSIZE);
      }
      n = min2(flen1, BUFSIZE);
    }
    if( n == 0 ) break;
    if( sendBytes(baseport, n, buffer1, v64format) != 0 ) {
      fflush(stdout);
      fprintf(stderr, "\n%s: Timeout while transmitting data to the Doctor V64!\n", arg);
      fprintf(stderr, "Please check cables and connection.\n");
#ifdef LINUX
spperror();
#endif
      exit(1);
    }
    bytessent += n;


status(bytessent,flen1);


    fflush(stdout);
    flen1 -= n;
  }
  printf("\n");
  return 0;
}

int receiveBytesToFile(int baseport, FILE *file1, long flen1, int v64format)
{
  printf("%8ld bytes (%ld.%02ld Mbit) will be transmitted.\n", flen1, flen1>>17, ((flen1*100)>>17)%100);
  fflush(stdout);
  for( ;; ) {
    unsigned int n;
    static long bytesreceived=0;

    n = min2(flen1, BUFSIZE);
    if( n <= 0 ) break;
    if( receiveBytes(baseport, n, buffer1, v64format) != 0 ) {
      fflush(stdout);
      fprintf(stderr, "\n%s: Timeout while receiving data from the Doctor V64!\n", arg);
      fprintf(stderr, "Please check cables and connection.\n");
#ifdef LINUX
spperror();
#endif
      exit(1);
    }
    bytesreceived += n;


status(bytesreceived,flen1);
/*    printf("\r%8ld Bytes received.", bytesreceived);
*/

    fflush(stdout);
    if( fwrite(buffer1, 1, n, file1) != n ) {
      fprintf(stderr, "\n%s: Error writing to output file!\n", arg);
      exit(1);
    }
    flen1 -= n;
  }
  printf("\n");
  return 0;
}

int sendFile(unsigned int port, char *fname1, long slen, int swapped)
{
  FILE *file1;
  long flen1;
  int v64format;

  fflush(stdout);
  if( (file1 = fopen(fname1, "rb")) == NULL ) {
    fprintf(stderr, "%s: Could not open file '%s' for reading.\n", arg, fname1);
    exit(1);
  }
  if( fread(buffer1, 1, 12, file1) != 12 ) {
    fprintf(stderr, "%s: File too short to be transmitted!\n", arg);
    exit(1);
  }
  fseek(file1, 0, SEEK_END);
  flen1 = ftell(file1);
  fseek(file1, 0, SEEK_SET);
  if( swapped == -1 ) {
    swapped = guessByteOrder((unsigned char *)buffer1);
  }
  if( (swapped == 0) || (swapped == 1) ) {
//    printf("The image '%s' is in %s format.\n", fname1, (swapped==0) ? "original (not swapped)" : "V64 (byte-swapped)");
  } else {
    if( (flen1 == 262144) && (slen == -1) ) {
      printf("The file '%s' appears to be a V64 BIOS file, OK.\n", fname1);
      swapped = 1;
    } else {
      if( swapped != -1 ) {
        swapped &= 1;
        printf("WARNING: I am not 100%% certain, but the image '%s' appears to be in %s format.\n", fname1, (swapped==0) ? "original (not swapped)" : "V64 (byte-swapped)");
      } else {
        fflush(stdout);
        fprintf(stderr, "%s: Could not detect type of image '%s', use commandline to override.\n", arg, fname1);
        exit(1);
      }
    }
  }
  if( slen != -1 ) {
    if( slen == -2 ) {
      if( biosVersion < 0x0130 ) {
        slen = (((flen1-1) >>23) +1 ) <<23;
      } else if( biosVersion < 0x0133 ) {
        slen = ( max2( ((flen1-1) >>18) +1, 5) ) <<18;
      } else {
        slen = (((flen1-1) >>18) +1 ) <<18;
      }
    }
  } else {
    slen = flen1;
  }
  v64format = swapped ^ 1;
  initCommunication(port);
  if( sendUploadHeader(port, fname1, slen) != 0 ) {
    fflush(stdout);
    fprintf(stderr, "%s: Error sending transfer headers to the Doctor V64!\n", arg);
    fprintf(stderr, "Please check cables and connection.\n");
#ifdef LINUX
spperror();
#endif
    exit(1);
  }
  sendBytesFromFile(port, file1, slen, v64format);
  fclose(file1);
  printf("\nFile transmitted successfully!\n");
  return 0;
}

int receiveFile(unsigned int port, char *fname1, long slen, int swapped)
{
  FILE *file1;
  int v64format;

  if( (file1 = fopen(fname1, "wb")) == NULL ) {
    fflush(stdout);
    fprintf(stderr, "%s: Could not open file '%s' for writing.\n", arg, fname1);
    exit(1);
  }
  if( swapped == -1 ) swapped = 1;
  printf("The image '%s' will be saved in %s format.\n", fname1, (swapped==0) ? "original (not swapped)" : "V64 (byte-swapped)");
  v64format = swapped^1;
  initCommunication(port);
  if( sendDownloadHeader(port, fname1, &slen) != 0 ) {
    fflush(stdout);
    fprintf(stderr, "%s: Error sending transfer headers to the Doctor V64!\n", arg);
    fprintf(stderr, "Please check cables and connection.\n");
#ifdef LINUX
spperror();
#endif
    exit(1);
  }
  receiveBytesToFile(port, file1, slen, v64format);
  fclose(file1);
  printf("\nFile transmitted successfully!\n");
  return 0;
}


int transfer(int mod)
{
  char fname1[256];
  long slen=-1;
  int swapped=-1;
  char romParameters[MAX_STRING_LENGTH];
//    int i;
//    char *c;
//    char tmpstr[MAX_STRING_LENGTH];
mode=mod;

    biosVersion = 0x0133;
    
    memset(romParameters, 0, sizeof(romParameters));

/*
        if( strstr(c, "sen") == c ) {
          mode = MODE_SEND;
          continue;
        }
        if( strstr(c, "rec") == c ) {
          mode = MODE_RECEIVE;
          continue;
        }
        if( strstr(c, "int") == c ) {
          mode = MODE_INTERACTIVE;
          continue;
        }
        if( strstr(c, "bios") == c ) {
          char t2[MAX_STRING_LENGTH];
          char *d=t2;
          c += strlen("bios");

          while( *c != '\0' ) {
            if( isdigit(*c) ) *d++ = *c;
            c++;
          }
          *d = '\0';
          if( sscanf(t2, "%x", &biosVersion) != 1 ) usage();
          if( (biosVersion < 0x0100) || (biosVersion > 0x0999) ) {
            fprintf(stderr, "%s: Please specify the BIOS version with three digits, e.g. -bios122.\n", arg);
            exit(1);
          }
          continue;
        }
        if( *c=='o' || *c=='s' ) {
          if( *(c+1) != '\0' ) usage();
          if( *c=='o' ) {
            swapped = 0;
          } else {
            swapped = 1;
          }
        } else
        if( *c=='p' ) {
          if( (strlen(c)<2) || (strlen(c)>BUFSIZE) ) usage();
          if( (sscanf(c+1, "%x%s", &port, buffer1) != 1) ) usage();
        } else
        if( *c=='l' ) {
          if( *(c+1) == '\0' ) {
            slen = -2;
            continue;
          }
          if( strlen(c)>BUFSIZE ) usage();
          if( sscanf(c+1, "%ld%s", &slen, buffer1) != 1 ) {
            if( sscanf(c+1, "%ldM%s", &slen, buffer1) == 1 ) {
              slen <<= 17;
            } else usage();
          }
        } else if( (*c == 'a') && (*(c+1) == '\0') ) {
          int j, k, l, p=0;
          for( j=i+1; j<argc; j++ ) {
            l = strlen(argv[j]);
            for( k=0; (k<l) && (p<MAX_STRING_LENGTH-1); k++, p++ ) {
              romParameters[p] = argv[j][k];
            }
            if( (j < argc-1) && (p<MAX_STRING_LENGTH-1) ) romParameters[p++] = ' ';
          }
          romParameters[HEADER_NAME_LEN-1] = '\0';
          break;
        } else usage();
      } else {
        if( fname1 == NULL ) {
          fname1 = argv[i];
        } else {
          usage();
        }
      }
    }
  }
*/

strcpy(fname1,cmd1);

  if( mode == MODE_UNDEF ) mode = MODE_SEND;
  if( mode == MODE_INTERACTIVE ) {
    if( fname1 != NULL ) usage();
  } else {
    if( (mode != MODE_SEND) || (romParameters[0] == '\0') ) {
      if( fname1 == NULL ) usage();
    }
  }

  port = getParPort(port);
  if( port == 0 ) {
    fflush(stdout);
    fprintf(stderr, "%s: Could not detect a parallel port, use commandline to specify one.\n", arg);
    exit(1);
  }
  printf("Using parallel port at %X; ", port);
  if( biosVersion >= 0x0130 ) {
    printf("BIOS 1.3%c extended mode.\n", (biosVersion >= 0x0133) ? '3' : '0');
  } else {
    printf("compatibility mode for pre 1.30 BIOSes.\n");
  }
  fflush(stdout);


start=time(0);



  if( mode == MODE_INTERACTIVE ) {
    interactiveMode(port);
  } else if( mode == MODE_SEND ) {
    if( fname1 != NULL ) {
      sendFile(port, fname1, slen, swapped);
    }
    if( romParameters[0] != '\0' ) {
      if( biosVersion < 0x0130 ) {
        fflush(stdout);
        fprintf(stderr, "%s: Argument sending not available with BIOSes earlier than 1.30!\n", arg);
        exit(1);
      }
      printf("Sending arguments...");
      fflush(stdout);
      if( (sendHeaderExtSend(port, HEADER_NAME_OFFSET, HEADER_NAME_LEN) != 0) ||
          (sendBytes(port, HEADER_NAME_LEN, romParameters, 1) != 0 ) ) {
        fprintf(stderr, "\n%s: Could not send arguments, make sure you have BIOS 1.30 or later.\n", arg);
        exit(1);
      }
      printf("done.\n");
    }
  } else {  
  mode = MODE_RECEIVE ;
    receiveFile(port, fname1, slen, swapped);
  }
  return 0;
}
