/*
    This file is part of the source of talk64 V1.5,
    a PC - DoctorV64 communications program
    Copyright (C) 1997  Andreas Sterbenz (stan@sbox.tu-graz.ac.at)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "global.h"
#include "v64proto.h"
#include "parsys.h"     /* definition of 'system-dependent' functions */

void initCommunication(unsigned int port)
{
  int i;
  for( i=0; i<SYNC_MAX_TRY; i++ ) {
    if( syncHeader(port) == 0 ) break;
  }
  if( i>=SYNC_MAX_TRY ) {
    fflush(stdout);
    fprintf(stderr, "%s: Could not establish communication with the Doctor V64!\n", progname);
    fprintf(stderr, "Verify that the V64 is powered on, ready to receive, and that the cable is ok!\n");
    exit(1);
  }
}

int sendBytes(unsigned int baseport, unsigned int len, char data[], int swap)
{
  long maxwait;
  unsigned int i;

#ifdef DEBUG_NOCOMM
  return 0;
#endif
  for( i=0; i<len; i++ ) {
    maxwait = SEND_MAX_WAIT;
    if( (in1byte(baseport+2) & 1) == 0 ) {      /* check ~strobe */
      while( ((in1byte(baseport+2) & 2) != 0) && maxwait-- ) ;  /* wait for */
      if( maxwait <= 0 ) return 1;                        /* auto feed == 0 */
      out1byte(baseport, data[i^swap]);         /* write data    */
      out1byte(baseport+2, 5);                  /* ~strobe = 1   */
    } else {
      while( ((in1byte(baseport+2) & 2) == 0) && maxwait-- ) ;  /* wait for */
      if( maxwait <= 0 ) return 1;                        /* auto feed == 1 */
      out1byte(baseport, data[i^swap]);         /* write data    */
      out1byte(baseport+2, 4);                  /* ~strobe = 0   */
    }
  }
  return 0;
}

int receiveBytes(unsigned int baseport, unsigned int len, char data[], int swap)
{
  int i;
  long maxwait;
  unsigned char c;

#ifdef DEBUG_NOCOMM
  for( i=0; i<len; i++ ) {
    data[i] = rand();
  }
  return 0;
#endif
  for( i=0; i<len; i++ ) {
    out1byte(baseport, REC_HIGH_NIBBLE);
    maxwait = REC_MAX_WAIT;
    while( ((in1byte(baseport+1) & 0x80) == 0) && maxwait-- ) ; /* wait for ~busy=1 */
    if( maxwait <= 0 ) return len-i;
    c = (in1byte(baseport+1) >> 3) & 0x0f; /* ~ack, pe, slct, ~error */
    out1byte(baseport, REC_LOW_NIBBLE);
    maxwait = REC_MAX_WAIT;
    while( ((in1byte(baseport+1) & 0x80) != 0) && maxwait-- ) ; /* wait for ~busy=0 */
    if( maxwait <= 0 ) return len-i;
    c |= (in1byte(baseport+1) << 1) & 0xf0; /* ~ack, pe, slct, ~error */
    data[i^swap] = c;
  }
  if( i != 0 ) {
    out1byte(baseport, REC_HIGH_NIBBLE);
  }
  return 0;
}

int checkSync(unsigned int baseport)
{
  int i, j;

#ifdef DEBUG_NOCOMM
  return 0;
#endif
  for( i=0; i<SYNC_MAX_CNT; i++ ) {
    if( ((in1byte(baseport+2) & 3) == 3) || ((in1byte(baseport+2) & 3) == 0) ) {
      out1byte(baseport, 0);                        /* ~strobe, auto feed */
      for( j=0; j<SYNC_MAX_CNT; j++ ) {
        if( (in1byte(baseport+1) & 0x80) == 0 ) {   /* wait for ~busy=0 */
          return 0;
        }
      }
      return 1;
    }
  }
  return 1;
}

int syncHeader(unsigned int baseport)
{
  int i=0;

#ifdef DEBUG_NOCOMM
  return 0;
#endif
  out1byte(baseport, 0);                        /* data = 00000000    */
  out1byte(baseport+2, 4);                      /* ~strobe=0          */
  while( i<SYNC_MAX_CNT ) {
    if( (in1byte(baseport+2) & 8) == 0 ) {      /* wait for select=0  */
      out1byte(baseport, 0xaa);                 /* data = 10101010    */
      out1byte(baseport+2, 0);                  /* ~strobe=0, ~init=0 */
      while( i<SYNC_MAX_CNT ) {
        if( (in1byte(baseport+2) & 8) != 0 ) {  /* wait for select=1  */
          out1byte(baseport+2, 4);              /* ~strobe=0          */
          while( i<SYNC_MAX_CNT ) {
            if( (in1byte(baseport+2) & 8) == 0 ) { /* w for select=0  */
              out1byte(baseport, 0x55);         /* data = 01010101    */
              out1byte(baseport+2, 0);          /* ~strobe=0, ~init=0 */
              while( i<SYNC_MAX_CNT ) {
                if( (in1byte(baseport+2) & 8) != 0 ) {  /* w select=1 */
                  out1byte(baseport+2, 4);      /* ~strobe=0          */
                  while( i<SYNC_MAX_CNT ) {
                    if( (in1byte(baseport+2) & 8) == 0) { /* select=0 */
                      return 0;
                    }
                    i++;
                  }
                }
                i++;
              }
            }
            i++;
          }
        }
        i++;
      }
      i++;
    }
    i++;
  }
  out1byte(baseport+2, 4);
  return 1;
}

int sendFilename(unsigned int baseport, char name[])
{
  int i;
  char *c;
  char mname[12];

  memset(mname, ' ', 11);
  c = max2(strrchr(name, '/'), strrchr(name, '\\'));
  if( c == NULL ) {
    c = name;
  } else {
    c++;
  }
  for( i=0; i<8 && *c!='.' && *c!='\0' ; i++, c++ ) mname[i] = toupper(*c);
  c = strrchr(c, '.');
  if( c != NULL ) {
    c++;
    for( i=8; i<11 && *c!='\0'; i++, c++ ) mname[i] = toupper(*c);
  }
  return sendBytes(baseport, 11, mname, 0);
}

int sendFileLen(unsigned int baseport, long len)
{
  char lenbuffer[4];

  lenbuffer[0] = (len    );
  lenbuffer[1] = (len>> 8);
  lenbuffer[2] = (len>>16);
  lenbuffer[3] = (len>>24);
  return sendBytes(baseport, 4, lenbuffer, 0);
}

int sendUploadHeader(unsigned int baseport, char name[], long len)
{
  static char protocolId[] = "GD6R\1";

  if( sendBytes(baseport, strlen(protocolId), protocolId, 0) != 0 ) return 1;
  if( sendFileLen(baseport, len) != 0 ) return 1;
  if( sendFilename(baseport, name) != 0 ) return 1;
  return 0;
}

int sendDownloadHeader(unsigned int baseport, char name[], long *len)
{
  static char protocolId[] = "GD6W";
  unsigned char recbuffer[15];

  if( sendBytes(baseport, strlen(protocolId), protocolId, 0) != 0 ) return 1;
  if( sendFilename(baseport, name) != 0 ) return 1;
  if( checkSync(baseport) != 0 ) return 1;
  if( receiveBytes(baseport, 1, (char *)recbuffer, 0) != 0 ) return 1;
  if( recbuffer[0] != 1 ) {
    fflush(stdout);
    fprintf(stderr, "Multi file download not supported (%d files)!\n", recbuffer[0]);
    exit(1);
  }
  if( receiveBytes(baseport, 15, (char *)recbuffer, 0) != 0 ) return 1;
  *len = (long)recbuffer[0] | ((long)recbuffer[1] << 8) | ((long)recbuffer[2] << 16) | ((long)recbuffer[3] << 24);
  return 0;
}

int sendHeaderExtRec(unsigned int baseport, long offset, long len)
{
  static char protocolId[] = "GD6ER";
  unsigned char buffer2[8];

  LONG2BYTES(offset, &buffer2[0], 0);
  if( len >= 0 ) {
    LONG2BYTES(len, &buffer2[4], 0);
    if( syncHeader(baseport) != 0 ) return 1;
  } else {
    LONG2BYTES(-len, &buffer2[4], 0);
  }
  if( sendBytes(baseport, strlen(protocolId), protocolId, 0) != 0 ) return 1;
  if( sendBytes(baseport, 8, (char *)buffer2, 0) != 0 ) return 1;
  if( checkSync(baseport) != 0 ) return 1;
  if( receiveBytes(baseport, 1, (char *)buffer2, 0) != 0 ) return 1;
#ifndef DEBUG_NOCOMM
  if( buffer2[0] != 0xa0 ) return 1;
#endif
  return 0;
}

int sendHeaderExtSend(unsigned int baseport, long offset, long len)
{
  static char protocolId[] = "GD6EW";
  unsigned char buffer2[8];

  LONG2BYTES(offset, &buffer2[0], 0);
  LONG2BYTES(len, &buffer2[4], 0);
  if( syncHeader(baseport) != 0 ) return 1;
  if( sendBytes(baseport, strlen(protocolId), protocolId, 0) != 0 ) return 1;
  if( sendBytes(baseport, 8, (char *)buffer2, 0) != 0 ) return 1;
  if( checkSync(baseport) != 0 ) return 1;
  if( receiveBytes(baseport, 1, (char *)buffer2, 0) != 0 ) return 1;
#ifndef DEBUG_NOCOMM
  if( buffer2[0] != 0xa1 ) return 1;
#endif
  return 0;
}

int sendHeaderExtPrint(unsigned int baseport, char line, char column)
{
  static char protocolId[] = "GD6EP";
  unsigned char buffer2[2];

  buffer2[0] = line;
  buffer2[1] = column;
  if( syncHeader(baseport) != 0 ) return 1;
  if( sendBytes(baseport, strlen(protocolId), protocolId, 0) != 0 ) return 1;
  if( sendBytes(baseport, 2, (char *)buffer2, 0) != 0 ) return 1;
  if( checkSync(baseport) != 0 ) return 1;
  if( receiveBytes(baseport, 1, (char *)buffer2, 0) != 0 ) return 1;
#ifndef DEBUG_NOCOMM
  if( buffer2[0] != 0xa3 ) return 1;
#endif
  return 0;
}

int sendHeaderExtCall(unsigned int baseport, int address)
{
  static char protocolId[] = "GD6EC";
  unsigned char buffer2[2];

  buffer2[0] = address >> 8;
  buffer2[1] = address & 0xff;
  if( syncHeader(baseport) != 0 ) return 1;
  if( sendBytes(baseport, strlen(protocolId), protocolId, 0) != 0 ) return 1;
  if( sendBytes(baseport, 2, (char *)buffer2, 0) != 0 ) return 1;
  if( checkSync(baseport) != 0 ) return 1;
  if( receiveBytes(baseport, 1, (char *)buffer2, 0) != 0 ) return 1;
#ifndef DEBUG_NOCOMM
  if( buffer2[0] != 0xa2 ) return 1;
#endif
  return 0;
}

